// -*- C++ -*-

// Scheduler.h
//
// Register and run different processes
// for Audio Pattern Recognition Language (aprl)
// dpwe@media.mit.edu 1995feb02
//

// I have to declare the Scheduler before the SchedProc to be able to 
// declare Scheduler methods as friends of SchedProc.  So a simple 
// forward declaration of SchedProc, the base class of objects with 
// an Update method that are executed by the Scheduler

#ifndef _SCHEDULER_H
#define _SCHEDULER_H

class SchedProc;

// MAPpable function type applied to a process, taking pointer to private data
typedef void SchedProcFn(SchedProc *d, void *dpfdata);

class SchedProcNode {	// Node for the list of SchedProcs we will define
public:
    SchedProcNode *next;
    SchedProc 	 *data;

    SchedProcNode(SchedProc *dp, SchedProcNode *nx = NULL) 
        { data = dp; next = nx; }
    ~SchedProcNode(void)	
        { };
};

class SchedProcList  {
private:
    SchedProcNode *head;
public:
    SchedProcList(void)	
        { head = 0; }
    ~SchedProcList(void);
    void insert(SchedProc *proc);	// splice into list by comparison
    int  remove(SchedProc *proc);	// find node matching proc, remove
    void apply(SchedProcFn *dpf, void *dpfdata);     // map dpf to each node
};

class Scheduler : private SchedProcList {	// Maintain a list of procs
                                                // and Update them in order.
public:
    Scheduler(void) : SchedProcList()	
        { };
    ~Scheduler(void)			
        { };
    
    void Update(double time);		// call update on all the tasks
    int Enschedule(SchedProc *proc);	// add an event to the schedule
    int Deschedule(SchedProc *proc);	// remove event from schedule
    int Reschedule(SchedProc *proc);	// change sequence of this proc
};

class SchedProc {	// base class for processes that are run by
			// the scheduler i.e. have a sequence priority
private:
    static int numberRegistered;	// universal counter of elements made
    double sequence;	// the particular sequence precedence of this process
    Scheduler *sched;	// where we are scheduled
//    int isScheduled;	// flag that this process is already in a schedule    
    friend int Scheduler::Enschedule(SchedProc *); // these modify the isSched
    friend int Scheduler::Deschedule(SchedProc *); //   field.
public:
    SchedProc(void)      
        { sequence = numberRegistered; ++numberRegistered; sched = 0; };
    virtual ~SchedProc(void)	
        { if(sched)	sched->Deschedule(this); };
    double Sequence(void) const	
        { return sequence; };
    int SetSequence(double newseq)
        { sequence = newseq;  if(sched != NULL)  sched->Reschedule(this); 
          fprintf(stderr, "<SP#0x%lx>: new seq is %.1f\n", this, sequence); 
           return 1; };
    virtual int Update(double time) = 0;  // do processing through till time?
    int IsScheduled(void) const
        { return sched != 0; }
};

// Magnitude comparator functions for SchedProc's allow them to be 
// sorted by execution sequence number in class Scheduler
inline int operator== (SchedProc &a, SchedProc &b)
    {  return a.Sequence() == b.Sequence();  }
inline int operator<  (SchedProc &a, SchedProc &b)
    {  return a.Sequence() < b.Sequence();  }
inline int operator>  (SchedProc &a, SchedProc &b)
    {  return a.Sequence() > b.Sequence();  }

#endif  /* _SCHEDULER_H */
