/*
 * Copyright 1994 Massachusetts Institute of Technology
 *
 * This software may be used for research purposes only. By prior
 * agreement with Lexicon, Inc. no part of this software may be
 * sold for profit or used in any commercial product, and the details
 * of the algorithms used may not be published or publicly presented.
 */
/*
 * Zero crossing detector. This works sequentially, rather than
 * processing a buffer of samples. Samples are processed by calling
 * xing_proc(), and valid zero crossings are returned via a callback
 * function. Call xing_query_params() to setup the parameters
 * and then setup the callback function. Zero crossings are defined
 * as the first positive sample after a negative sample, with the
 * condition that the positive swing must exceed the envelope.
 *
 * Bill Gardner, January, 1994.
 */
#include    <stdio.h>
#include    <math.h>

extern "C" {
#ifdef BILLG
#include    "misc.h"
#include    "userface.h"
#else 
#define TRUE (1)
#define FALSE (0)  
#endif /* BILLG */
};

#include    "Xing.h"

#define DEBUG	FALSE
#define GLOBAL

extern double tc_to_gain(double tc, long samprate);
extern long samprate;

#define PI  3.14159265

/*
 * Callback function called with valid xing:
 *
 * accept_fn(long delta)
 *
 * delta is the time in samples since the last zero crossing.
 */

#define STATE_NEG   0	/* samples are negative, waiting for xing */
#define STATE_XING  1	/* have xing, wait for peak to validate */
#define STATE_POS   2	/* found valid xing, wait for negative samples */

/*
 * 1-pole, 1-zero AC-coupling highpass filter normalized
 * for unit magnitude at Nyquist frequency.
 */
double xing::hpf_proc(double samp)
{
double zin, out;
    zin = samp * (1 + hpf_gain) / 2 + hpf_z * hpf_gain;
    out = zin - hpf_z;
    hpf_z = zin;
    return out;
}

/*
 * 1 pole lowpass normalized for unit magnitude at DC.
 */
double xing::lpf_proc(double samp)
{
double zin;
    zin = samp + lpf_gain * lpf_z;
    lpf_z = zin;
    return zin * (1 - lpf_gain);
}

inline double fc_to_gain(double fc, long samprate)
{
    return 1 - 2 * PI * fc / samprate;
}

void xing::Set_hpf_fc(double f)
{
    hpf_fc = f;
    hpf_gain = fc_to_gain(hpf_fc,samprate);
}

void xing::Set_lpf_fc(double f)
{
    lpf_fc = f;
    lpf_gain = fc_to_gain(lpf_fc,samprate);
}

void xing::Set_decay_tc(double t)
{
    decay_tc = t;
    decay = tc_to_gain(decay_tc, samprate);
}

/*
 * Initialize the zero crossing finder.
 */
xing::xing(long srate)
{
    samprate  = srate;
    state = STATE_NEG;
    env = 0;
    count = 0;
    xing_c = 0;
    hpf_z = 0;
    lpf_z = 0;
    // external controls to reasonable defaults
    lpf_flag = TRUE;
    hpf_flag = TRUE;
    Set_hpf_fc(100.0);
    Set_lpf_fc(200.0);
    Set_decay_tc(16.0);
}

xing::~xing(void)
{
}

void xing::default_params()
{
    hpf_flag = TRUE;
    hpf_fc = 100;
    lpf_flag = TRUE;
    lpf_fc = 200;
    decay_tc = 16;
}

/*
 * Query user for zero crossing finder parameters.
 */
void xing::query_params()
{
#ifdef BILLG
    if (hpf_flag = yes_no("highpass filter input",hpf_flag)) {
	getdouble("HPF cutoff frequency (Hz)",&hpf_fc);
    }
    if (lpf_flag = yes_no("lowpass filter input",lpf_flag)) {
	getdouble("LPF cutoff frequency (Hz)",&lpf_fc);
    }
    getdouble("envelope decay time constant (msec)",&decay_tc);
#endif /* BILLG */
}

/*
 * Process a sample. Tricky state machine. When a zero-crossing is
 * detected, call the "acceptance" callback function.
 */
void xing::proc(short samp)
{
double fsamp = samp;
    /*
     * Run highpass and lowpass filter.
     */
    if (hpf_flag) fsamp = hpf_proc(fsamp);
    if (lpf_flag) fsamp = lpf_proc(fsamp);
    /*
     * Common to all states: increment counter and run the
     * envelope follower.
     */
    count++;
    if (fsamp > env) env = fsamp;
    else env *= decay;

    switch (state) {
    case STATE_NEG:
	/*
	 * Check for positive sample.
	 */
	if (fsamp <= 0) break;
	/*
	 * Remember zero crossing and change state.
	 */
	xing_c = count;
	state = STATE_XING;
	/*
	 * Falls through!
	 */
    case STATE_XING:
	/*
	 * Check for sample to exceed envelope.
	 */
	if (fsamp >= env) {
	    /*
	     * Call the acceptance function, and reset count and state.
	     */
	    count -= xing_c;
	    state = STATE_POS;
	    accept_fn(xing_c);
	    break;
	}
	/*
	 * Falls through!
	 */
    case STATE_POS:
	/*
	 * We've accepted a zero-crossing, now wait for samples to
	 * go negative again. If we've fallen through from STATE_XING,
	 * then we're just checking for a negative sample.
	 */
	if (fsamp < 0) state = STATE_NEG;
	break;
    }
}
