#!/usr/local/bin/wish
#
# thislif.tcl
# 
# Interface to the thislIR program
# 
# 1999jan31 dpwe@icsi.berkeley.edu
# $Header: /homes/dpwe/tmp/thisl/thislguipkg-1999may05/thislgui/RCS/thislif.tcl,v 1.3 1999/09/17 13:23:38 dpwe Exp $

#
# Configuration
#

# Should we use speech input?
set useAudioFrontend 1

# Should we play back real audio files from SoundDir using rvplayer, 
# or can we use sndplay on uncompressed audio found with bn_file?
set useRealAudio 1

# Should we use the Thomson NLP processing on the recognizer output?
set useNLP 1

# Should we use the Abbot ASR chain (else local special case)
# Guess...
if {[file exists "/u/drspeech/data/bn/src/simplebn"]} {
    set useAbbot 0
} else {
    set useAbbot 1
}

# If we're not using realaudio, can we use bn_file to find the files?
# (or are they just in audiodir?)
if {[catch {set wavfile [exec bn_file a960521]}]} {
    # bn_file failed
    set useBnFile 0
} else {
    # bn_file worked.  If we can find the audio, maybe we can avoid RA
    set useBnFile 1
    if {[file exists $wavfile]} {
	set useRealAudio 0
    }
}

#
# Set up basic directory paths etc.
#

package require Dpwe_Utilfns

# Load the other thislir interface commands
set ScriptDir [file dirname [info script]]
# ScriptDir is also used to find the image files (in the images subdir)

# basedir is parent of directory containing script
#set basedir ".."
#set basedir [file dirname $ScriptDir]
# Now its the same place as the scripts
set basedir $ScriptDir

# ThislDir is path to the root of the standard thislIR demo directory
# either from env(THISL_DIR) or relative to basedir
set ThislDir [getenv "THISL_DIR" [file join $basedir "thisldemo"]]

# Later thisl demos have data in subdir - try to guess if we have such
set ThislDataDir [file join $ThislDir "data"]
if {![file exist $ThislDataDir]} {
    set ThislDataDir $ThislDir
}

set speech_arch [SpeechArch]
set ThislArch $speech_arch
set BinDir   [getenv "THISL_BIN_DIR" [file join $ThislDir $ThislArch "bin"]]

if {![file exist $BinDir]} {
    # New thisl uses different arch names
    set ThislArch [join [list [exec uname -p] "-" [exec uname -m] "-" [exec uname -s]] ""]
    # .. with an exception for solaris
    regsub {SunOS} $ThislArch "solaris" ThislArch
    # So, try again
    set BinDir [getenv "THISL_BIN_DIR" [file join $ThislDir $ThislArch "bin"]]
}

if {![file exist $BinDir]} {
    # Not an arch-specific bin dir?
    Warn "Couldn't find arch-specific bin dir '$BinDir'"
    set BinDir [file join $ThislDir "bin"]
}

set IndexDir [getenv "THISL_INDEX_DIR" [file join $ThislDataDir "index"]]
set AudioDir [getenv "THISL_AUDIO_DIR" [file join $ThislDataDir "audio"]]
# maybe prefer local copies of soundfiles
set ScratchDir [getenv "THISL_SCRATCH_DIR" "/scratch/thisl"]
set AudioDirLocal [file join $ScratchDir "audio"]
if {[file exists $AudioDirLocal]} {
    Warn "Using local soundfiles from $AudioDirLocal"
    set AudioDir $AudioDirLocal
}
set SrtDir  [getenv "THISL_SRT_DIR" [file join $ThislDataDir "srt"]]
set SrtDirLocal [file join $ScratchDir "srt"]
if {[file exists $SrtDirLocal]} {
    Warn "Using local srt files from $SrtDirLocal"
    set SrtDir $SrtDirLocal
}
if {![file exist $SrtDir]} {
    Warn "Couldn't find SRT directory $SrtDir"
    set SrtDir ""
}

# The thislIR binary
set thislIR     [file join $BinDir "thislIR"]
# Constants for ThislIR
set b 0.5
set k1 1.0
set maxout 20

# Maybe start the thislIR server
if {[catch {set thislIRvsn [exec $thislIR -version 2> /dev/null | grep thislIR | awk {{print $2}}]} msg]} {
    if {[string match "*no such*" msg]} {
	Error "Cannot execute thislIR binary '$thislIR'"
    }
    #puts "trying $thislIR: got '$msg'"
    # vsn 0.2 exits abnormally with -version
    set thislIRvsn 0.2
}

set useServer 0
# Later versions of thislIR return stories by times, not by index
set timesReturned 0

if {$thislIRvsn > 0.3} {
    # This is a server version
    set useServer 1
    set timesReturned 1
}

# Files in the index directory
set episodelist [file join $IndexDir "episodelist"]
set doclist     [file join $IndexDir "doclist"]
set stoplist    [file join $IndexDir "stoplist"]
set nostemlist  [file join $IndexDir "nostem"]
set unigramList [file join $IndexDir "bn96-65p-1-2.unigram.list"]
set indexFile   [file join $IndexDir "index"]

# Sometimes, the index file is compressed
if {![file exists $indexFile]} {
    append indexFile ".gz"
}

# Where temporary query/response files are written
set workingdir  $IndexDir

set thislsrvpid ""

if {$useServer} {
    # values from THISLdemo.pm of 0.3 vsn
    set thislPort 21997
    set thislHost [exec hostname]
    set thislClient [file join $BinDir "thislClient"]
    # Is there a server already running?
    if {![catch {set running [exec ps -ef | grep $thislIR | grep -v grep]}]} {
	Warn "Server already running (?): $running"
    } else {
	# Start the server
	set thislsrvpid [exec $thislIR -server $thislPort -index $indexFile -maxout $maxout -b $b -K1 $k1 -qtf -quiet &]
    }
}

# Maybe have path to thomson NLP parser
set ParserDir [getenv "THISL_PARSER_DIR" "/u/dpwe/tmp/thisl/thomson/nathalie/bin"]
# If runparser can't be found, turn OFF useNLP
if {$useNLP} {
    if {![file executable "$ParserDir/runparser"]} {
	Warn "Cannot run $ParserDir/runparser - NLP skipped"
	set useNLP 0
    }
}

#
# Set up the audio_frontend input and ASR interface
#

if {$useAudioFrontend} {
    package require AudioFrontend
    package require Dpwe_Rdprocmgr
    package require Dpweutils_tcl

    #
    # Configure one or other of the speech recognizers
    #

    # Basename for lattice output files (if useNLP)
    set latticebase "/tmp/thisl-lat-[pid]"

    set samplerate 16000
    if {$useAbbot} {
	# Configure for ASR components from the AbbotDemo distribution
	set AbbotDir [getenv "ABBOT_DEMO_DIR" [file join $basedir "AbbotDemo"]]
	if {![file readable $AbbotDir]} {
	    Error "Cannot find AbbotDemo directory at $AbbotDir"
	} else {
	    set AbbotArch [exec [file join $AbbotDir "demoHostType"]]
	    set AbbotBin  [file join $AbbotDir $AbbotArch "bin"]
	    set AbbotEtc  [file join $AbbotDir "etc"]
	    set LM	$AbbotEtc/spring98-20k-64-64.bin
	    set dialect "UK" ;# or "UK", whose dict seems to crash Noway
	    if {$dialect == "UK"} {
		# UK dialect settings
		set WEIGHTS	$AbbotEtc/wsjcam0_plp_zap.wei
		set PHONES	$AbbotEtc/UKphone.beep
		set PRIORS	$AbbotEtc/UKpriors.beep
		set PRONS	$AbbotEtc/BEEP20k.dict
		set NPHONE	45
	    } else {
		# US dialect settings
		set WEIGHTS	$AbbotEtc/bn97_F0+F1.3.0_zap.wei
		set PHONES	$AbbotEtc/USphone.icsi
		set PRIORS	$AbbotEtc/USpriors.icsi
		set PRONS	$AbbotEtc/CMU20k.dict
		set NPHONE	54
	    }
	    set recogCmd "$AbbotBin/plp -online | $AbbotBin/rnnInputNorm -nobuffer | $AbbotBin/rnnForwardDemo $WEIGHTS - - | $AbbotBin/noway -demo -online -phone_models $PHONES -priors $PRIORS -dictionary $PRONS -trigram $LM -lna $NPHONE"
	    if {$useNLP} {
		# Need lattice outputs
		append recogCmd " -lattice $latticebase"
	    }
	}
    } else {
	# Don't use abbott - use our local Broadcast News net
	set simpledir "/u/drspeech/data/bn/src/simplebn"
	set recogCmd "rasta -M -F -k -S $samplerate -m 12 -n 13 -s 16 -w 32 -q 0 | rnnInputNorm -fpout -index8 -index8out -bigout -ninp 13 | qnsfwd ftr1_file=- ftr1_format=onlftr ftr1_width=13 ftr1_ftr_start=0 ftr1_ftr_count=13 ftr1_norm_file=$simpledir/plp12N-16k.norms window_extent=9 ftr1_window_len=9 init_weight_file=$simpledir/boot-plp12N-16k-70h-aI5+117i+2000h+54o.wts mlp3_input_size=117 mlp3_hidden_size=2000 mlp3_output_size=54 mlp3_output_type=softmax realtime=true activation_file=- activation_format=lna8 log_file=/dev/null debug=0 verbose=false"
	if {$useNLP} {
	    # Need lattice outputs - have to use noway
	    append recogCmd " | noway -demo -lna -phone_models $simpledir/CI54_BN98.align3.model.ICSI -priors $simpledir/CI54_BN98.align3.priors -dictionary $simpledir/bn98-20k.dct -trigram $simpledir/spring98-20k-64-64.bin -lattice $latticebase"
	} else {
	    # No lattice outputs - can use Chronos
	    append recogCmd " | chronos -phi $simpledir/CI54_BN98.align3.phi -dct $simpledir/bn98-20k.dct -lm $simpledir/spring98-20k-64-64.bin"
	}
    }

    puts "recogCmd=$recogCmd"

    #
    # Some other stuff needed for the audio_frontend interface
    #

    # Handler for audio samples back from audio frontend
    proc getAudio {pipe} {
	# Get samples from audio_frontend and write them to the 
	# forward-pass pipe
	global recogPipe
	set n [bfilecopy $pipe $recogPipe 2048]
	if {[eof $pipe]} {
	    rdprocClosePipe $pipe
	    #puts stderr "getAudio: closed ($n bytes in final block)"
	    flush $recogPipe
	}
    }

    # Handler for utterances read back from decoder
    proc getWords {pipe} {
	# Output from the decoder
	global useNLP
	set line [gets $pipe]
	if {$line == ""} {
	    Warn "getWords: skipping blank line"
	} else {
    	    Warn "getWords: $line"
	    if {[eof $pipe]} {
		rdprocClosePipe $pipe
		puts "Decode closed"
	    }
	    # Strip the leading "count" returned by noway
	    #regsub {^[0-9]* *} $line "" line
	    regexp {([0-9]*) *(.*)} $line all uttnum line
	    # Pass it to the backend (but not the first line)
	    if {![string match "noway*" $line]} {
		# Maybe filter through Thomson NLP
		if {$useNLP} {
		    set line [NLPparse $line]
		}
		QRYdo [string tolower $line]
	    }
	}
    }

    # Make sure pipes are off during set up (also, force load 
    # of rdprocMgr, before audio_frontend defines its own.
    # Make sure pipes are active
    rdprocDisable

    # Build the recognition chain
    # Do it BEFORE opening audio_frontend, so recognizer doesn't inherit
    # and hold open the /dev/dsp file handle, thereby preventing playback.
    set recogPipe [open "| $recogCmd" "w+"]
    fconfigure $recogPipe -translation binary -buffering none -blocking 0
    rdprocAdd $recogPipe "getWords $recogPipe"

}
# End of ASR input setup


#
# Now define a set of fns to handle the backend access to thislIR
# and audio output
#

proc ReadUnigrams {filename} {
    # Read a list of unigram probabilities into an array
    global uniLProbArray
    puts "ReadUnigrams: reading '$filename'..."
    set f [open $filename "r"]
    set count 0
    while {![eof $f]} {
	set l [gets $f]
	if {$l != ""} {
	    set uniLProbArray([lindex $l 0]) [lindex $l 1]
	    incr count
	    if {[expr $count % 1000] == 0} {
		puts -nonewline "$count "
		flush stdout
	    }
	}
    }
    close $f
    puts "\nReadUnigrams: got $count words from $filename"
}

# Set up the unigram array
resetArray uniLProbArray
#Warn "Skipping unigrams (for speed)"
ReadUnigrams $unigramList

# Threshold for excluding common words
set commonLProb -2.7

proc QNormalize {query} {
    # Apply all the normalization as in thisdemo.pm
    # Also, remove common words
    global uniLProbArray commonLProb
    set normstring ""
    foreach word $query {
	regsub -all {i\.e\.} $word " " word ;# Remove i.e.
	regsub -all {e\.g\.} $word " " word ;# Remove e.g.
	regsub -all {[-?,;:/()]} $word " " word ;# unwanted chars
	regsub -all {\.} $word ". " word ;# space after period
	regsub -all {([A-z_][A-z_]+)\.} $word {\1} word ;# remove period but not after abbrev
	if {[regexp {[A-Z][A-Z]+} $word]} {
	    # Word was all upper case - treat as abbreviation, break into 
	    # individual characters 
	    regsub -all {([A-Z])} $word {\1. } word
	}
	# Make whole thing lower case
	set word [string tolower $word]
	set uword [string toupper $word]
	# add on to result, if it's not too common
	if {$word != "" \
		&& ([array names uniLProbArray $uword] == "" \
		|| $uniLProbArray($uword) < $commonLProb) } {
	    if {$normstring != ""} {
		append normstring " "
	    }
	    append normstring $word
	}
    }
    # Strip repeated spaces
    regsub -all {  +} $normstring " " normstring
    return [string trim $normstring]
}

proc WriteQuery {filename query} {
    # Write a query file suitable for thislir
    set f [open $filename "w"]
    puts $f "<top>\n\n<num> Number: 1\n\n<desc> Description:"
    puts $f "$query"
    puts $f "\n</top>"
    close $f
}

proc ReadResults {filename} {
    # Parse a query results file from ThislIR
    set f [open $filename "r"]
    set rslt {}
    while {![eof $f]} {
	set l [string trim [gets $f]]
	if {$l != "" && [string range $l 0 0] != "#"} {
	    lassign $l num qry fileID cnt score qfile
	    lappend rslt [list $fileID $score]
	}
    }
    close $f
    return $rslt
}

proc DoQuery {query} {
    # Run thislir on the query
    global thislIR thislClient thislHost thislPort
    global workingdir indexFile b k1 maxout
    global resdisp srtdisp
    global tcl_platform
    global rootwin
    global useServer
    # Display the status
    afe_setstatus "running thislir..."; update
    # Set up constants
    set qfile [file join $workingdir "query"]
    set rfile [file join $workingdir "result"]
    set nquery [QNormalize $query]
    WriteQuery $qfile $nquery
    set err 0
    set rslt ""
    if {$tcl_platform(platform) != "macintosh"} {
	# Clear the existing results panel
	RESclear $resdisp
	# Set the pause cursor
	set ecursor [$rootwin cget -cursor]
	$rootwin configure -cursor watch
	update
	# Run the actual thislIR query
	# Are we using a server or not?
	if {$useServer} {
	    set cmd "$thislClient $thislHost $thislPort < $qfile > $rfile"
	} else {
	    set cmd "$thislIR -index $indexFile -queries $qfile -maxout $maxout -out $rfile -b $b -K1 $k1"
	}
	puts stderr "cmd=$cmd"
	set err [catch {eval exec $cmd} rslt]
	puts stderr "rslt=$rslt"
	# Restore the entry cursor
	$rootwin configure -cursor $ecursor
    }
    afe_setstatus "displaying hits..."; update
    # Read back the result
    puts "DoQuery: done ($err)"
    if {$err == 0} {
	set rslt [ReadResults $rfile]
	# Clear the current SRT display (prior to offering a new set)
	SRTclear $srtdisp
	# Display the query hits (will abort if user starts recording)
	RESfill $resdisp [list $nquery $rslt]
    }
    afe_setstatus "idle"; update
}

if {$useRealAudio} {
    proc playStart {epi st} {
	return [playStartRA $epi $st]
    }
    proc playStop {} {
	return [playStopRA]
    }
} else {
    # Not RA - use WAVfile access
    proc playStart {epi st} {
	return [playStartWAV $epi $st]
    }
    proc playStop {} {
	return [playStopWAV]
    }
}

set playpid ""
set afe_audiosrc ""

proc playStartWAV {episode start} {
    # Commence sound output for the named episode at the specified time
    global playpid AudioDir useBnFile
    global tcl_platform
    global afe_audiosrc
    playStop
    set rc 0
    if {$useBnFile} {
	set sndfile [exec bn_file $episode]
    } else {
	set sndfile [file join $AudioDir "$episode.au"]
    }
    if {$tcl_platform(platform) != "macintosh"} {
	if {$afe_audiosrc != ""} {$afe_audiosrc close}
	set playpid [exec sndplay $sndfile -k $start & ]
	set rc 1
    }
    return $rc
}

proc playStopWAV {} {
    # Terminate any currently-playing sound
    global playpid afe_audiosrc
    if {$playpid != ""} {
	catch {exec kill $playpid}
	set playpid ""
	if {$afe_audiosrc != ""} {$afe_audiosrc open}
    }
}

# Paths to needed binaries
# Assume rvplayer will be in the path
set rvplayer "rvplayer"
# Try arch-dep rmcut from THISLDemo tree first
set rmcut [file join $ThislDir "opt/pnserver/$speech_arch/bin/rmcut"]
if {![file exists $rmcut]} {
    # Failing that, try arch-indep
    set rmcut [file join $ThislDir "opt/pnserver/bin/rmcut"]
    if {![file exists $rmcut]} {
	set rmcut    "rmcut"
	Warn "No direct path to rmcut found - hope it's in the PATH"
    }
}

# Local file for 5 min cut-out pieces
set ratmpfile "/tmp/tmp.ra"

proc playStartRA {episode start} {
    # Play real audio segment
    global AudioDir
    global afe_audiosrc playpid
    global rvplayer rmcut ratmpfile
    playStop
    set rc 0
    set sndfile [file join $AudioDir "$episode.ra"]
    # There seems to be a skew (or a coarse granularity) in playing back 
    # real audio, so compensate
    set ra_precomp 1.2
    set start [expr $start - $ra_precomp]
    set cmd "$rmcut -i $sndfile -o $ratmpfile -s $start -e [expr $start+300.0]"
    if {[catch {eval exec $cmd} rslt]} {
	Warn "Error $rslt executing $cmd"
    } else {
	if {$afe_audiosrc != ""} {$afe_audiosrc close}
	set playpid [exec $rvplayer $ratmpfile &]
	set rc 1
    }
    return $rc
}
proc playStopRA {} {
    global afe_audiosrc playpid
    global rvplayer rmcut ratmpfile
    if {$playpid != ""} {
	set playpid ""
	set cmd "exec $rvplayer {}"
	if {$afe_audiosrc != ""} {
	    append cmd "; $afe_audiosrc open"
	}
	# Restore the button first, so it doesn't look ugly while we're blocked
	update
	# The idea is that if I can fork this off and run it in the 
	# background, it won't block the event loop while "exec rvplayer """
	# decides what to do.  But I think it blocks anyway.
	# We want to wait for the return from "rvplayer """ before 
	# reopening the record pipe, just to make sure that raplay
	# really has finised.
	#after 1 $cmd
	eval $cmd
	# This takes a couple of seconds to complete, but at least it will 
	# stop
    }
}

#
# Hook called just before exit to permit us to cleanup
#
proc thislif_cleanup {} {
    # We need to kill the server (if its running)
    global useServer thislsrvpid
    Warn "thislif_cleanup: useServer=$useServer thislsrvpid=$thislsrvpid"
    if {$useServer && $thislsrvpid != ""} {
	Warn "killing thislIR server on pid $thislsrvpid..."
	exec kill $thislsrvpid
    }
}

#
# Source the file to setup the interface to the Thomson NLP code
# 
if {$useNLP} {
    source [file join $ScriptDir "thislnlp.tcl"]
}


#
# Run the code to set up the graphical user interface and get things 
# happening
#
source [file join $ScriptDir "thislui.tcl"]

# End
