#
# thislui.tcl
#
# Tcl/Tk front-end to thislIR demo
#
# 1999jan28 dpwe@icsi.berkeley.edu  for thisldemo 0.2
# $Header: /homes/dpwe/tmp/thisl/thislguipkg-1999may05/thislgui/RCS/thislui.tcl,v 1.5 1999/09/17 13:23:12 dpwe Exp $

# This package should have already been required by thislif.tcl, but 
# include it here just in case this file is being sourced elsewhere
package require Dpwe_Utilfns

# Tcl7/Tcl8 compatibility
if {[string match "7.*" $tcl_version]} {
    # Tcl 7 - clumsy fonts, menus, no images in text
    set fontArray(helv24bi) "*helvetica-bold-o*--24-*"
    set fontArray(helv12)   "*helvetica-medium-r-*--12-*"
    set fontArray(helv12b)  "*helvetica-bold-r-*--12-*"
    set fontArray(cour10)   "*courier-medium-r-*--10-*"
    set isTcl7 1
} else {
    # Tcl 8 - nice
    set fontArray(helv24bi) "Helvetica 24 bold italic"
    set fontArray(helv12)   "Helvetica 12"
    set fontArray(helv12b)  "Helvetica 12 bold"
    set fontArray(cour10)   "Courier 10"
    set isTcl7 0
}

# Padding for frames
set pad 5
# Highlight color background
set highlightColor "LightBlue"
# Maximum for scaling of score bars
set scoremax 18.0

proc ParseSGMLhdr {hdr} {
    # Take an SGML header tag like:
    # <Section Type=FILLER S_time=3.613 E_time=28.160 ID=ea980107.2 >
    # and parse it up to set variables in the calling context.
    # (in this case, $Type, $S_time, $E_time, $ID)
    # Return a list of the variables set.
    set varnames {}
    if {[regexp {<([-A-z0-9_]*)  *(.*)>} $hdr all tag vals]} {
	# Assume it's param=val, so replace "=" with " " to 
	# make sure quoted vals are parsed OK by Tcl.
	regsub -all "=" $vals " " vals
	set param ""
	foreach el $vals {
	    if {$param == ""} {
		set param $el
	    } else {
		# 2nd of pair is val
		upvar $param $param
		set $param $el
		# Record that we set this param
		lappend varnames $param
		# Clear this param name since we've used it
		set param ""
	    }
	}
    }
    return $varnames
}

proc ReadSRTcl {episode} {
    # Read our tcl-preformatted version of the SRT-file data
    global SrtDir allShows
    set filename [file join $SrtDir "$episode.tcl"]
    if {[catch {source $filename} rslt]} {
	Warn "ReadSRTcl: error reading '$filename': $rslt"
	set data ""
    } else {
	# Ensure all known programs are in showlist
	set epiprog [lindex $data 1]
	if {[lsearch $allShows $epiprog] == -1} {
	    ShowlistAdd $epiprog
	    # adds it to global allShows too
	}
    }
    return $data
}

proc WriteSRTcl {episode data} {
    # Write out preformatted SRT data in our tcl format
    global SrtDir
    set filename [file join $SrtDir "$episode.tcl"]
    if {![file writable $SrtDir]} {
	Warn "WriteSRTcl: cannot write into $SrtDir"
	return 0
    } elseif {[file exists $filename] && ![file writable $filename]} {
	Warn "WriteSRTcl: cannot overwrite $filename"
    } else {
	set f [open $filename "w"]
	puts -nonewline $f "set data \{"
	puts -nonewline $f $data
	puts $f "\}"
	close $f
	Warn "WriteSRTcl: '$filename' written"
    }
}

proc ReadSRT {episode {bbcFmt 0}} {
    # Parse an SRT file (text tagged with times and broken into 
    # segments).  Return a list of segment transcripts, which are 
    # each labdur lists i.e. {start dur word} {start dur word} ...
    # Actual return is of form:
    # {file prog date {sectID type start end {{start dur wd} ...}}}
    # Optional bbcFmt flag means that filenames are in bbc format,
    # not BN format - used to extract show date
    global SrtDir allShows
    set filename [file join $SrtDir "$episode.srt"]
    set f [Open $filename "r"]
    if {$f == ""} {
	Warn "ReadSRT: can't read '$filename'"
	return ""
    }
    set epifile ""
    set segid ""
    set segwords {}
    set allsegs {}
    set epidone 0
    while {![eof $f]} {
	set l [gets $f]
	# Optimization for <Word>..</Word> lines
	if {[string match "<Word*/Word>" $l]} {
	    #<Word S_time=3.613 E_time=3.789>on</Word>
	    # modified to accept <Word ...><s></Word>
	    regexp {=([^ ]*)[^=]*=([^>]*)>(.[^<]*)<} $l all S_time E_time text
	    if {$text == "<s>"} {
		set text "."
	    } elseif {$text == "<SIL>"} {
		set text "/"
	    }
	    lappend segwords [list $S_time $E_time $text]
	} else {
	    set l [string trim $l]
	    if {[string range $l 0 0] != "#"} {
		while {$l != ""} {
#		    if {$epidone} {
#		        Warn "ReadSRT($filename): got '$l' after episode end"
#		    }
		    # Run through this line, treating as seq of ... and <...>
		    if {[regexp { *([^<]*) *<([^>]*)> *(.*)} $l all text sgmltag rest] == 0} {
			Error "Cannot parse '$l' as sgml-tagged text"
		    }
		    # OK, we got it as some chunk of SGML
		    # store the part we haven't parsed
		    set l $rest
		    # Figure out the SGML tag and act accordingly
		    switch -glob $sgmltag {
			# First, are we closing an existing tag?
			"/Word" {
			    # Just pick up the current vals of these vars
			    lappend segwords [list $S_time $E_time $text]
			    # Clear them for debug
			    unset S_time E_time text
			} 
			"/Section" {
			    lappend allsegs [list $segid $segtype $segstt $segend \
						  $segwords]
			    # Clear the variables
			    unset segtype segstt segend
			    set segid ""
			    set segwords {}
			}
			"/Episode" {
			    # It's the end of the episode - expect no more!
			    set epidone 1
			}
			"Episode *" {
			    # Episode header
			    set attrs [ParseSGMLhdr "<${sgmltag}>"]
			    # Assume no previous episode header
			    if {$epifile != ""} {
				Error "ReadSRT($filename): eh <$sgmltag> after Episode $epifile"
			    }
			    # Check for required attributes
			    if { !( [lsearch $attrs "Filename"] > -1 \
				    && [lsearch $attrs "Program"] > -1 \
				    && [lsearch $attrs "Version_Date"] > -1 ) } {
				Error "ReadSRT($filename): no Filename/Program/Version_Date in <$sgmltag>"
			    }
			    set epifile $Filename
			    if {[lsearch $attrs "Channel"] > -1} {
				set epiprog "$Channel: $Program"
			    } else {
				set epiprog $Program
			    }
			    # Ensure all known programs are in showlist
			    if {[lsearch $allShows $epiprog] == -1} {
				ShowlistAdd $epiprog
				# adds it to global allShows too
			    }
				
			    #set epidate $Version_Date
			    # Date is encoded in filename
			    if {$bbcFmt} {
				if {![regexp {(..)(.)(..)(..)} $epifile all show yr mo da]} {
				    Error "ReadSRT($filename): can't map '$epifile' to BBC date"
				}
				# Restore missing decade (Y2K!!!!)
				if {$yr > 5} {
				    set yr "9$yr"
				} else {
				    set yr "0$yr"
				}
			    } else {
				if {![regexp {..(..)(..)(..)} $epifile all yr mo da]} {
				    Error "ReadSRT($filename): can't map '$epifile' to BN date"
				}
			    }
			    if {[string range $mo 0 0] == "0"} {
				set mo [string range $mo 1 1]
			    }
			    set mname [lindex ". jan feb mar apr may jun jul aug sep oct nov dec" $mo]
			    if {[string range $yr 0 0] < 5} {
				set yr "20$yr"
			    } else {
				set yr "19$yr"
			    }
			    set epidate $yr$mname$da
			    # Clear the vars
			    eval unset $attrs
			}
			"Section *" {
			    # Section marker
			    set attrs [ParseSGMLhdr "<${sgmltag}>"]
			    # Assume not nested
			    if {$segid != ""} {
				Error "ReadSRT($filename): <${sgmltag}> inside seg $segid"
			    }
			    # Check for required attributes
			    if { !( [lsearch $attrs "Type"] > -1 \
				    && [lsearch $attrs "ID"] > -1 \
				    && [lsearch $attrs "S_time"] > -1 \
				    && [lsearch $attrs "E_time"] > -1 ) } {
				Error "ReadSRT($filename): no Type/ID/S_time/E_time in <$sgmltag>"
			    }
			    set segid $ID
			    set segtype $Type
			    set segstt $S_time
			    set segend $E_time
			    # Clear the vars
			    eval unset $attrs
			}
			"Word *" {
			    # Word - just let S_time and E_time hang around
			    set attrs [ParseSGMLhdr "<${sgmltag}>"]
			}
			"*" {
			    Error "ReadSRT($filename): unknown SGML tag <$sgmltag>"
			}
		    }
		}
	    }
	}
    }
    close $f
    if {$segwords != ""} {
	Error "ReadSRT($filename): unclosed seg $segid at EOF"
    }
    puts stderr "ReadSRT: read '$filename'"
    return [list $epifile $epiprog $epidate $allsegs]
}

# Image bits
set imageset {}

proc SetupImages {} {
    # Setup a string of images for including in the index display
    global imageset ScriptDir
    set ImageDir [file join $ScriptDir "images"]
    set imageset {}
    set colors {steelblue yellow3 orange red}
    set ncolors [llength $colors]
    set nimages 11
    for {set i 0} {$i < $nimages} {incr i} {
	set clr [lindex $colors [expr $ncolors*$i/$nimages]]
	set img [image create bitmap -background white -foreground $clr \
		       -file [file join $ImageDir "bar$i.xbm"] \
		       -maskfile [file join $ImageDir "barmask.xbm"]]
	lappend imageset $img
    }
    return $imageset
}

proc txtof {path} {
    # Convert a path name to the index of the text widget
    # (same for both RES and SRT displays)
    return $path.txf.text
}

# How many lines in the results summary display?
set RESlines 10
# How many 'columns' in the results summary display?
set REScols 100

proc RESmake {path} {
    # Construct a pane to display the queries
    global pad highlightColor RESlines REScols
    global fontArray
    frame $path -borderwidth 2 -relief groove
    # Title line
    set ti [frame $path.title -relief flat]
    set ll {}
    lappend ll [label $ti.l1 -text "Results for:" -width 14]
    lappend ll [label $ti.l2 -text "" -width 60 -relief sunken -anchor w -bd 1]
    eval pack $ll -side left -fill y -expand 0
    pack $ti -side top -fill x -expand 0 -padx $pad -pady $pad
    # Display window
    frame $path.txf -relief flat
    set hdr [text $path.txf.hdr \
                -width 70 -height 1 \
		-borderwidth 2 -relief flat -cursor {}]
    pack $hdr -side top -fill x
    set txt [text [txtof $path] -yscrollcommand "$path.txf.scroll set" \
                -width $REScols -height $RESlines -wrap none \
		-borderwidth 2 -relief sunken -bg white \
		-state disabled -cursor {}]
    set scr [scrollbar $path.txf.scroll -command "$txt yview"]
    pack $scr -side right -fill y
    pack $txt -expand yes -fill both
    pack $path.txf -side top -padx $pad -pady $pad -fill both -expand 1
    # 5 columns: relevance, show, date, time, text
    set  tabs ".5c 1.5c 8c 11c 13c"
    # Set up aligned header
    $hdr configure -tabs $tabs
    $hdr tag configure plain -font $fontArray(helv12)
    $hdr insert end "\t\tProgram\tDate\tOffset\tContext" plain
    $hdr config -state disabled
    # Setup display window
    $txt configure -tabs $tabs
    $txt tag configure plain -font $fontArray(helv12)
    $txt tag configure bold  -font $fontArray(helv12b)
    $txt tag configure highlight -background $highlightColor
}

proc GetContext {seg qry {width 15}} {
    # $seg is the text of a segment; $qry is a list of query terms.
    # Return the word-index of the first occurrence of one of the 
    # query words in the seg, and a string of 15 words surrounding that 
    # point (1/3 before and 2/3 after)
    set pos -1
    foreach wrd $qry {
	if {[set wpos [lsearch -exact $seg $wrd]] > -1} {
	    if {$pos == -1 || $wpos < $pos} {
		set pos $wpos
	    }
	}
    }
    set bpos [max 0 [expr $pos - $width/3]]
    set epos [expr $bpos+$width-1]
    set slen [llength $seg]
    if {$epos > $slen} {
	set xs [expr $epos - $slen]
	set bpos [max 0 [expr $bpos - $xs]]
	set epos $slen
    }
    set rslt [list $pos [lrange $seg $bpos $epos]]
    return $rslt
}

proc GetSegwds {worddata stime etime} {
    # Subselect a {start end word} list to return only words within
    # the indicated time limits; but also return how many words were 
    # skipped since the start of the list
    set nskipwds 0
    set segwds {}
    foreach wdrcd $worddata {
	lassign $wdrcd stt end wd
	if {$end > $etime} {
	    break
	} elseif {$stt < $stime} {
	    incr nskipwds
	} else {
	    lappend segwds $wd
	}
    }
    return [list $nskipwds $segwds]
}

# Global array holds all the parsed SRT files we read, keyed by fileID
if {[info vars srtArray] == ""} {
    resetArray srtArray
}

proc RESclear {path} {
    # Blank out the results panel
    set ti $path.title
    $ti.l2 configure -text ""
    set txt [txtof $path]
    $txt config -state normal
    $txt delete 1.0 end
    $txt config -state disabled
}

proc FilterOK {prog date} {
    # Check the specified program & date against the filter criteria
    # Return 1 if ok, 0 to skip
    # First the program name
    global startDateWidget endDateWidget
    set OKshows [ShowlistGet]
    if {[lsearch $OKshows $prog] == -1} {
	Warn "Excluding show '$prog' - not in '$OKshows'"
	return 0
    }
    # Now the date
    # $date is probably in format 1999mar03 - convert to secs
    set showsecs [clock scan "[string range $date 4 6] [string range $date 7 8], [string range $date 0 3]"]
    set begsecs [clock scan [DateGet $startDateWidget]]
    set endsecs [clock scan [DateGet $endDateWidget]]
    if {$showsecs < $begsecs || $showsecs > $endsecs} {
	Warn "Excluding date $date ($showsecs) from $begsecs..$endsecs"
	return 0
    }
    return 1
}

proc SortHitsCmp {i1 i2} {
    # Sort command for hits.  Just compares real values as first el
    set v1 [lindex $i1 0]
    set v2 [lindex $i2 0]
    if {$v1 < $v2} {
	return -1
    } elseif {$v1 > $v2} {
	return 1
    } else {
	return 0
    }
}

proc MergeStories {hits} {
    # Go through the hits returned by thislIR and merge those 
    # which come from overlapping segments.  Re-estimate scores 
    # for merged stories and re-sort.
    # Only works for "timesReturned"-style hits
    set parsedhits {}
    foreach hit $hits {
	lassign $hit filesegID score
	regexp {([-A-z0-9_]*)-([.0-9]*)-([.0-9]*)-([.0-9]*)} $filesegID all fileID stime ctime etime
	lappend parsedhits [list $score $stime $etime $ctime $fileID]
    }
    set mergedhits {}
    set i 0
    while {$i < [llength $parsedhits]} {
	set basehit [lindex $parsedhits $i]
	incr i
	lassign $basehit score stime etime ctime fileID
	set nmerge 1
	set j $i
	foreach remhit [lrange $parsedhits $j e] {
	    lassign $remhit rscore rstime retime rctime rfileID
	    if {$fileID == $rfileID \
		    && ($rstime <= $etime || $retime >= $stime)} {
		# Do the merge!
		Warn "Merging $fileID $stime-$etime & $rstime-$retime"
		set stime [min $stime $rstime ]
		set etime [max $etime $retime ]
		set score [expr $score + $rscore ]
		incr nmerge
		# Remove this story from the list
		set parsedhits [lreplace $parsedhits $j $j]
	    } else {
		# Look at next story
		incr j
	    }
	}
	# Boil down the overall score
	set score [expr $score/$nmerge]
	# Build new list
	lappend mergedhits [list $score $stime $etime $ctime $fileID]
    }
    # Sort by new scores
    set mergedhits [lsort -decreasing -command SortHitsCmp $mergedhits]
    # Now return to hits format
    set outhits {}
    foreach hit $mergedhits {
	lassign $hit score stime etime ctime fileID
	lappend outhits [list $fileID-$stime-$ctime-$etime $score]
    }
    return $outhits
}

# This state variable is changed by audio_frontend when the user 
# starts recording another query
set afe_state "idle"

proc RESfill {path data} {
    # Set up a bunch of queries.  $path is from RESdisp
    # $data is {words {{hit1file hit1score} {hit2file hit2score}}}
    global srtArray rootwin imageset scoremax isTcl7
    global afe_state
    global timesReturned
    set nimages [llength $imageset]
    # Set the pause cursor
    set ecursor [$rootwin cget -cursor]
#    $rootwin configure -cursor watch
    update
    #
    lassign $data query hits
    set ti $path.title
    $ti.l2 configure -text $query
    set txt [txtof $path]
    $txt config -state normal
    $txt delete 1.0 end

    # Maybe merge the hits
    if {$timesReturned} {
	set hits [MergeStories $hits]
    }

    # Max number to display, after merging and filtering
    set maxhitnum 10
    set hitnum 0    
    foreach hit $hits {
	# Abort the (slow) reading of SRT files if the user has 
	# started recording another query
	if {$afe_state != "idle"} {break}
	# Also, we return a lot of hits to allow for filtering, but 
	# only display some of them
	if {$hitnum >= $maxhitnum} {break}

	# Process this hit
	lassign $hit filesegID score
	# Is this a thislIR-0.2 return (<file>.<index>) or
	# a thislIR-0.5 return (<file>-<start>-<mid>-<end>)
	# For now, we'll equate this to BN or BBC data too
	set bbcFmt 0
	if {$timesReturned} {
	    regexp {([-A-z0-9_]*)-([.0-9]*)-([.0-9]*)-([.0-9]*)} $filesegID all fileID stime ctime etime
	    # For later on, use the filesegID as just the filename
	    set filesegID $fileID
	    # Note that this is bbc fmt filenames
	    set bbcFmt 1
	} else {
	    regexp {([-A-z0-9_]*).([0-9]*)} $filesegID all fileID seg
	    # Flag that stime/etime etc not set
	    set stime ""
	    # Strip word offset from filesegID
	    set filesegID $fileID.$seg
	}
	if {[array names srtArray $fileID] == ""} {
	    # Don't have this SRT file - load it, maybe save back tcl fmt
	    if {[set srtArray($fileID) [ReadSRTcl $fileID]] == ""} {
		# Read of tcl format failed, try original, maybe write back
		if {[set srtArray($fileID) [ReadSRT $fileID $bbcFmt]] != ""} {
		    WriteSRTcl $fileID $srtArray($fileID)
		}
	    }
	}
	# May have failed - so skip if so
	if {[llength $srtArray($fileID)] > 0} {
	    lassign $srtArray($fileID) filename progname date segs
	    # Check against our filter criteria
	    if {[FilterOK $progname $date]} {
		incr hitnum
		# Search through to the right seg
		foreach segdata $segs {
		    lassign $segdata segid type start end worddata
		    if {$segid == $filesegID} break
		}
		if {$segid != $filesegID} {
		    Error "couldn't find seg $filesegID in file $fileID"
		}
		# Use limit times from seg desc if not already set
		if {$stime == ""} {
		    set stime $start
		    set etime $end
		    set ctime [expr ($start+$end)/2.0]
		}
		# Setup context
		#set segwds [lmap $worddata "nth 2"]
		lassign [GetSegwds $worddata $stime $etime] skipwds segwds
		lassign [GetContext $segwds $query] conpos conwds
		set conpos [expr $skipwds+$conpos]
		set filesegIDpos "${filesegID}-$stime-$etime:$conpos"
		# Figure image from score
		set imgix [expr int($nimages*$score/$scoremax)]
		if {$imgix < 0} {set imgix 0}
		if {$imgix >= $nimages} {set imgix [expr $nimages-1]}
		# Put tagged tab to start tagged region
		$txt insert end "\t" [list plain $filesegIDpos]
		if {$isTcl7} {
		    # No embedded images in tcl7 text widgets
		    set im $txt.h$hitnum
		    if {[info command $im] == ""} {
			# Create embeddable label widgets just once
			label $im -borderwidth 0
		    }
		    $im configure -image [lindex $imageset $imgix]
		    $txt window create end -window $im
		} else {
		    $txt image create end -image [lindex $imageset $imgix]
		}
		$txt insert end "\t$progname\t$date\t[secs2time $stime]\t" \
		  [list plain $filesegIDpos]
		foreach wd $conwds {
		    set style plain
		    if {[lsearch -exact $query $wd] > -1} {
			set style bold
		    }
		    $txt insert end "$wd " [list $style $filesegIDpos]
		}
		$txt insert end "\n" plain
		$txt tag bind $filesegIDpos <1> "RESclick $filesegIDpos \"$query\""
		# Display it now
		update
	    }
	}
    }
    $txt config -state disabled
    $rootwin configure -cursor $ecursor
}

proc RESclick {filesegIDpos qwords} {
    # User clicked on a query - display it
    global srtdisp resdisp
    # Split id:pos into pieces
#    regexp {([^:]*):(.*)} $filesegIDpos all filesegID conpos
    # Highlight the query
    set txt [txtof $resdisp]
    $txt tag remove highlight 1.0 end
    $txt tag add highlight $filesegIDpos.first $filesegIDpos.last
    # Scroll to segment
    SRTview $srtdisp $filesegIDpos
    SRTemph $srtdisp $qwords
}

proc Segidpos2index {segidpos} {
    # Convert a filename.seg+wordbase:word ID to a line.char format
    global currentSegixs srtdisp
    set txt [txtof $srtdisp]
#    regexp {([^:]*):(.*)} $segidpos all segid wdpos
    regexp {([^\.\+]*)(\.([^:]*))?(\+([0-9]*))?:(.*)} $segidpos all fileID dummy1 segix dummy2 wdbase wdpos
    set segid "${fileID}$dummy1$dummy2"
    if {[set lineno [lsearch -exact $currentSegixs $segid]] < 0} {
	Error "couldn't find seg $segid (from $segidpos) in currentSegixs ($currentSegixs)"
    }
    # Retrieve the text for that line
    incr lineno
    set segwds [$txt get $lineno.0 $lineno.end]
    # Figure offset to nth word
    if {$wdbase == ""} {set wdbase 0}
    set chrs [string len [lrange $segwds 0 [expr $wdpos - $wdbase]]
    # Return the index, including separator following the word
    return $lineno.[expr $chrs+1]
}

proc RESSegidpos2index {segidpos} {
    # Convert a filename.seg-time-time:word ID to a line.char format
    global currentSegixs srtdisp
    set txt [txtof $srtdisp]
#    regexp {([^:]*):(.*)} $segidpos all segid wdpos
    regexp {([^\.\+]*)(\.([^:]*))?-([.0-9]*)-([.0-9]*):(.*)} $segidpos all fileID dummy1 segix stime etime wdpos
    set segid "${fileID}$dummy1"
    set lineno ""
    set wdbase ""
    set linecount 0
    foreach segix $currentSegixs {
	lassign [split $segix "+"] thissegid wdbase
	if {$thissegid == $segid} {
	    if {$wdbase > $wdpos} {
		# Found pos after this word
		break
	    }
	    # Else record this line as our best bet
	    set lineno $linecount
	}
	incr linecount
    }
#    if {[set lineno [lsearch -exact $currentSegixs $segid]] < 0}
    if {$lineno == ""} {
	Error "couldn't find seg $segid (from $segidpos) in currentSegixs ($currentSegixs)"
    }
    # Retrieve the text for that line
    incr lineno
    set segwds [$txt get $lineno.0 $lineno.end]
    # Figure offset to nth word
    if {$wdbase == ""} {set wdbase 0}
    set chrs [string len [lrange $segwds 0 [expr $wdpos - $wdbase]]]
    # Return the index, including separator following the word
    return $lineno.[expr $chrs+1]
}

proc Index2segidpos {index} {
    # Convert a line.char index into filename.seg(+word):word ID
    global currentSegixs srtdisp
    set txt [txtof $srtdisp]
    regexp {([^\.]*)\.(.*)} $index all lineno chrno
    set segid [lindex $currentSegixs [expr $lineno-1]]
    # Break fileID[.seg][+wdbase] into parts
    set wdbase ""
    regexp {([^\.\+]*)(\.([^:]*))?(\+([0-9]*))?} $segid all fileID dummy1 segix dummy2 wdbase
    # Get the words
    set segwds [$txt get $lineno.0 $lineno.end]
    # Count the words in the prefix
    set wds [llength [string range $segwds 0 $chrno]]
    # discount the first word (i.e. the time tag)
    incr wds -2
    # .. but add on the base of this seg
    if {$wdbase == ""} {
	set wdbase 0
    }
    incr wds $wdbase
    set segidpos $segid:$wds
    return $segidpos
}

proc Segidpos2wdinfo {segidpos} {
    # Get the actual time etc. corresponding to segidpos
    global srtArray
    if {![regexp {([^\.\+]*)(\.([0-9]*))?(\+([0-9]*))?:(.*)} $segidpos all fileID dummy1 segix dummy2 wdbase wdpos]} {
	Error "couldn't parse segidpos $segidpos"
    }
#    regexp {([^\.]*)(\.([^:]*))?-([.0-9]*)-([.0-9]*):(.*)} $segidpos all fileID dummy segix stime etime wdpos
    set filesegID "${fileID}$dummy1"
    # Have to look at the SRT data
    lassign $srtArray($fileID) filename progname date segs
    # Search through to the right seg
    foreach segdata $segs {
	lassign $segdata segid type start end worddata
	if {$segid == $filesegID} break
    }
    if {$segid != $filesegID} {
	Error "couldn't find seg $filesegID in file $fileID"
    }
    # Now worddata is the array
    set wrdrcd [lindex $worddata $wdpos]
    return $wrdrcd
}


proc SRTview {srtdisp segidpos} {
    # Make the SRTdisp display the specified seg
    global srtArray
    # The .segix part is not present in bbc format files, so allow skip
    regexp {([^\.]*)(\.([^:]*))?-([.0-9]*)-([.0-9]*):(.*)} $segidpos all fileID dummy segix stime etime wdpos
    SRTfill $srtdisp $srtArray($fileID)
    # Now scroll
    set txt [txtof $srtdisp]
    $txt see [RESSegidpos2index $segidpos]
}

# Global holds the stop button on the SRT display for stopping playback
set SRTstopBtn ""

proc SRTmake {path} {
    # $txt is a widget path; create some nested widgets to 
    # display SRT data
    global pad SRTstopBtn
    global fontArray
    frame $path -borderwidth 2 -relief groove
    # Title line
    set ti0 [frame $path.title -relief flat]
    set ti [frame $ti0.left -relief flat]
    set ll {}
    lappend ll [label $ti.l3 -text "Program:"]
    lappend ll [label $ti.l4 -text "" -width 24 -relief sunken -anchor w -bd 1]
    lappend ll [label $ti.l5 -text "Date:"]
    lappend ll [label $ti.l6 -text "" -width 11 -relief sunken -anchor w -bd 1]
#    lappend ll [label $ti.l7 -text "Time:"]
#    lappend ll [label $ti.l8 -text "" -width 8 -relief sunken -anchor w -bd 1]
    lappend ll [label $ti.l1 -text "File:"]
    lappend ll [label $ti.l2 -text "" -width 10 -relief sunken -anchor w -bd 1]
    eval pack $ll -side left -fill y -expand 0
    pack $ti -side left -fill none -expand 0
    set SRTstopBtn [button $ti0.stop -text "Stop playback" -state disabled -command "\$SRTstopBtn config -state disabled; playStop"]
    pack $ti0.stop -side right -fill none -expand 0
    pack $ti0 -side top -fill x -expand 0 -padx $pad -pady $pad
    # Display window
    frame $path.txf -relief flat
    set txt [text [txtof $path] -yscrollcommand "$path.txf.scroll set" \
                -width 70 -height 10 -wrap word \
		-borderwidth 2 -relief sunken -bg white \
		-state disabled -cursor {}]
    set scr [scrollbar $path.txf.scroll -command "$txt yview"]
    pack $scr -side right -fill y
    pack $txt -expand yes -fill both
    pack $path.txf -side top -padx $pad -pady $pad -fill both -expand 1
    # Set up text
    set indent "1.5c"
    $txt configure -spacing1 0.05i -tabs $indent
    $txt tag configure id   -font $fontArray(cour10) \
      -lmargin1 0 -lmargin2 $indent
    $txt tag configure text -font $fontArray(helv12) \
      -lmargin1 0 -lmargin2 $indent    
    $txt tag configure emph -font $fontArray(helv12b) \
      -lmargin1 0 -lmargin2 $indent    

}

proc secs2time {secs} {
    # Return a second count into a nicely-formatted "3m45s" string
    set mins [expr int($secs/60)]
    set rsecs [expr int($secs-60*$mins)]
#    return [format "%dm%02ds" $mins $rsecs]
    return [format "%02d:%02d" $mins $rsecs]
}

proc wdclick {wdtag w x y} {
    # Callback when one of the words is clicked
    # index within the text widget:
    global SRTstopBtn
    set index [$w index @$x,$y]
    # hence segidpos
    set segidpos [Index2segidpos $index]
    # hence word info
    set wdinfo [Segidpos2wdinfo $segidpos]
    # hence play command
    lassign $wdinfo btime etime wrd
    regexp {([^\.\+]*)(\.([^:]*))?(\+([0-9]*))?:(.*)} $segidpos all fileID dummy1 segix dummy2 wdbase wdpos
    puts "play $fileID $btime ($wrd)"
    if {[playStart $fileID $btime]} {
	# Enable the stop button
	$SRTstopBtn config -state normal
    }
}

proc wdunclick {wdtag w x y} {
    puts "stop"
    playStop
}

proc SRTclear {path} {
    # Clear the existing (old) contents of the SRT window
    set ti $path.title.left
    $ti.l2 configure -text ""
    $ti.l4 configure -text ""
    $ti.l6 configure -text ""
    set txt [txtof $path]
    $txt config -state normal
    $txt delete 1.0 end
    $txt config -state disabled    
}

proc SRTfill {path data} {
    # $path is an SRT structure created by SRTdisp; $data is SRT data 
    # from ReadSRT; display the data in the window
    # $data is like:
    # {file prog date {sectID type start end {{start dur wd} ...}}}
    global currentSegixs
    lassign $data fname prog date segdata
    # Setup title fields
    set ti $path.title.left
    # If we're already showing this data, skip it
    if {$fname == [$ti.l2 cget -text]} {
	return
    }    
    $ti.l2 configure -text $fname
    $ti.l4 configure -text $prog
    $ti.l6 configure -text $date
    # Stuff the display
    set txt [txtof $path]
    $txt config -state normal
    $txt delete 1.0 end
    set currentSegixs {}
    set lastbrk 1
    foreach seg $segdata {
	lassign $seg sectID type segstart segend words
	set wdcnt 0
	# Build a list to map segIDs to line numbers
	# Create a unique tag for this block
	# Semaphore if following a break
	foreach wd $words {
	    if {$lastbrk} {
		set wdtag "$sectID+$wdcnt"
		set wdtags [list "text" $wdtag]
		lappend currentSegixs $wdtag
		set stime [lindex $wd 0]
		$txt insert end "[secs2time $stime]\t" [list id $wdtag]
		# Debug - put tag
		#$txt insert end "$wdtag\t" [list id $wdtag]
		set lastbrk 0
		$txt tag bind $wdtag <1> "wdclick $wdtag %W %x %y"
		# $txt tag bind $wdtag <ButtonRelease-1> "wdunclick $wdtag %W %x %y"
	    }
	    set wrd [lindex $wd 2]	    
	    if {$wrd == "/"} {
		# Para break on words
		$txt insert end "\n"
		set lastbrk 1
	    } else {
		$txt insert end "$wrd " $wdtags
	    }
	    incr wdcnt
	}
	# Para break
	$txt insert end "\n" text
	set lastbrk 1
    }
    $txt config -state disabled
}

# textSearch --
# Search for all instances of a given string in a text widget and
# apply a given tag to each instance found.
#
# Arguments:
# w -		The window in which to search.  Must be a text widget.
# string -	The string to search for.  The search is done using
#		exact matching only;  no special characters.
# tag -		Tag to apply to each instance of a matching string.

proc textSearch {w string tag} {
    if {$string == ""} {
	return
    }
    set cur 1.0
    while 1 {
	set cur [$w search -count length $string $cur end]
	if {$cur == ""} {
	    break
	}
	$w tag add $tag $cur "$cur + $length char"
	set cur [$w index "$cur + $length char"]
    }
}

proc SRTemph {path words} {
    # Emphasize words from the list in the current SRT display
    set txt [txtof $path]
    $txt tag remove "emph" 0.0 end
    foreach w $words {
	textSearch $txt $w "emph"
    }
}


proc DateSelmake {path name} {
    # Create a series of popups for date selection
    frame $path -borderwidth 2 -relief flat
    # Title at top
    set ti [frame $path.title -relief flat]
    set la [label $ti.l1 -text $name -width 12]
    #pack $ti -side top -fill x -expand 0 -padx 5 -pady 5
    #pack $la -side left -fill y -expand 0
    # Series of popups underneath
    #set pfr [frame $path.popfr -relief flat]
    set pfr $ti
    set months {January February March April May June July August September October November December}
#    set momenu [menu $histmenu]
#    foreach mo $months {
#	$momenu insert end command -label $mo -command "puts $mo"
#    }
#    set monbut [menubutton $pfr.month]
    set monbut $pfr.month
    set daybut $pfr.day
    set yrbut  $pfr.year
    set monmenu [eval tk_optionMenu $monbut mymonth$path $months]
    set days {}
    for {set i 1} {$i <= 31} {incr i} {
	if {$i < 10} {
	    set dd "0$i"
	} else {
	    set dd $i
	}
	lappend days $dd
    }
    set daymenu [eval tk_optionMenu $daybut myday$path $days]
    set yrs {}
    for {set i 1995} {$i <= 2004} {incr i} {lappend yrs $i}
    set yrmenu [eval tk_optionMenu $yrbut myyr$path $yrs]
    pack $la $yrbut $monbut $daybut -side left -fill y -expand 0
    pack $pfr -side top -fill y -expand 0 -padx 5 -pady 5
    $yrbut configure -width 4
    $monbut configure -width 9
    $daybut configure -width 2

    return $path
}

proc menuSearch {path val} {
    # find an entry in a menu by its value
    set max [$path index end]
    for {set i 0} {$i <= $max} {incr i} {
	set mval [$path entrycget $i -label]
	if {$mval == $val} {
	    return $i
	}
    }
    return -1
}


proc DateSet {path date} {
    # Set the date to reflect a given date
    set datesecs [clock scan $date]
    set ymd [clock format $datesecs -format "%Y %B %d"]
    set path $path.title
    set ymenu $path.year.menu
    set mmenu $path.month.menu
    set dmenu $path.day.menu
    $ymenu invoke [menuSearch $ymenu [lindex $ymd 0]]
    $mmenu invoke [menuSearch $mmenu [lindex $ymd 1]]
    $dmenu invoke [menuSearch $dmenu [lindex $ymd 2]]
}

proc DateGet {path} {
    # Return text form of date from selection, as January 01, 1999
    # (which is parseable by DateSet)
    set path $path.title
    set ymenub $path.year
    set mmenub $path.month
    set dmenub $path.day
    set yr  [$ymenub cget -text]
    set mon [$mmenub cget -text]
    set day [$dmenub cget -text]
    return "$mon $day, $yr"
}

#set allShows {"BBC1: Midnight News" "BBC1: One O'Clock News" "BBC1: 4pm News" "BBC1: 5pm News" "BBC1: Six O'Clock News" "BBC1: 8pm News" "BBC1: Nine O'Clock News"}
set allShows {}
# Maybe read from a file?
set showFile [file join $IndexDir "shows.txt"]
if {[file exist $showFile]} {
    set f [open $showFile "r"]
    set allShows [lsort [lwithout [split [read -nonewline $f] "\n"] ""]]
    close $f
}

set showListWidget ""

proc ShowlistMake {path name} {
    # Create a scrolling listbox to contain selectable shows
    global allShows showListWidget

    frame $path -relief flat -borderwidth 2
    
    set la [label $path.la -text $name]
    set lb [listbox $path.lb -height 4 -width 30 \
	    -selectmode multiple -exportselection 0\
	    -yscrollcommand "$path.scroll set"]
    # must have exportselection 0 else will get deselected by X selects
    set scr [scrollbar $path.scroll -command "$lb yview"]

    pack $la $lb $scr -side left -fill y -expand 0

    # Insert all the shows we know about, and select them all
    foreach show $allShows {
	$lb insert end $show
    }
    $lb selection set 0 end

    set showListWidget $path

    return $path
}

proc ShowlistGet {{all 0}} {
    # Return the list of shows from the Showlist.  
    # Normally returns just the selected shows, but will 
    # return all shows if all=1
    global showListWidget
    set lb $showListWidget.lb
    set max [$lb index end]
    set rslt {}
    if {$all} {
	for {set i 0} {$i < $max} {incr i} {
	    lappend rslt [$lb get $i]
	}
    } else {
	foreach i [$lb curselection] {
	    lappend rslt [$lb get $i]
	}
    }
    return $rslt
}

proc ShowlistAdd {entry} {
    # Append a new show name to the list
    global showListWidget allShows
    set lb $showListWidget.lb
    # Try to insert it alphabetically
    set insix "end"
    set current [ShowlistGet 1]
    for {set i 0} {$i < [llength $current]} {incr i} {
	if {$entry < [lindex $current $i]} {
	    set insix $i
	    break;
	}
    }
    $lb insert $insix $entry
    $lb selection set $insix
    # Make it visible too
    $lb see $insix
    # Add it to the global list if it's not already there
    if {[lsearch $allShows $entry] == -1} {
	lappend allShows $entry
    }
}

# Global record of the entry widget (for QRYdo)
set QRYent ""
proc QRYmake {path} {
    # Create the query panel in the specified path
    global pad highlightColor histmenu QRYent startDateWidget endDateWidget
    frame $path -borderwidth 2 -relief groove
    # Title line
    set ti [frame $path.title -relief flat]
    set la  [label $ti.l1 -text "Enter query:" -width 12]
    set QRYent [entry $ti.ent -width 60 -relief sunken \
	     -bg white -selectbackground $highlightColor -selectborderwidth 0]
    # -font "fixed"
    # History popdown menu
    set mnubut $ti.hist
    set histmenu $mnubut.mnu
    menubutton $mnubut -indicatoron 1 -menu $histmenu
    # -direction below (tcl8 only)
    menu $histmenu
    grid $mnubut -row 0 -column 1 -sticky ne
    pack $la -side left -fill y -expand 0
    pack $mnubut -side right -fill y -expand 0
    pack $QRYent -side left -fill both -expand 1
    pack $ti -side top -fill x -expand 0 -padx $pad -pady $pad
    # Add date selection
    set filtfr [frame $path.filter -relief flat]
    set filts [DateSelmake $filtfr.start "Start date:"]
    set filte [DateSelmake $filtfr.end "End date:"]
    DateSet $filte "December 31, 2004"
    pack $filts $filte -side top -fill y -expand 0
    pack $filtfr -side left -fill y -expand 0
    set showlist [ShowlistMake $path.shows "Programs:"]
    pack $showlist -side left -fill y -expand 0 -padx 30

    set startDateWidget $filts
    set endDateWidget $filte

    # Bindings
    # <Enter> initiates retrieval
    bind $QRYent <Return> "QRYdo \[$QRYent get\]"
    # focus gain makes whole contents be selected, so typing will replace
    bind $QRYent <FocusIn> "$QRYent selection range 0 end"
}

proc QRYdo {qry {addtohist 1}} {
    # Do a query based on the passed-in words
    global QRYent
    puts "QRYdo $qry"
    # Write into entry (in case from elsewhere)
    $QRYent delete 0 end
    $QRYent insert end $qry
    $QRYent selection range 0 end
    if {$addtohist} {
	# Add to history (not if invoked from hist)
	set histmenu [winfo parent $QRYent].hist.mnu
	$histmenu insert 0 command -label $qry -command "QRYdo \"$qry\" 0"
    }
    # snatch the focus from the qry panel
    #focus [winfo parent $QRYent]
    # Run the query
    DoQuery $qry
}

proc MNUmake {w} {
    # Make the menu bar in toplevel $w
    # (Menubar code from std menu demo)
    global tcl_platform isTcl7

    if {$isTcl7} {
	frame $w.menu -relief raised -borderwidth 2
	set menubuts ""
	set mf $w.menu.file.menu
	lappend menubuts [menubutton $w.menu.file -text "File" -menu $mf]
	set mo $w.menu.opts.menu
	lappend menubuts [menubutton $w.menu.opts -text "Options" -menu $mo]
    } else {
	# Tcl 8 - cleaner, integrated menus
	menu $w.menu -tearoff 0
	set mf $w.menu.file
	$w.menu add cascade -label "File" -menu $mf -underline 0
	set mo $w.menu.opts
	$w.menu add cascade -label "Options" -menu $mo -underline 0
    }
    menu $mf -tearoff 0
    $mf add command -label "Open..." -command {error "this is just a demo: no action has been defined for the \"Open...\" entry"}
    $mf add command -label "New" -command {error "this is just a demo: no action has been defined for the \"New\" entry"}
    $mf add command -label "Save" -command {error "this is just a demo: no action has been defined for the \"Save\" entry"}
    $mf add command -label "Save As..." -command {error "this is just a demo: no action has been defined for the \"Save As...\" entry"}
    $mf add separator
    $mf add command -label "Print Setup..." -command {error "this is just a demo: no action has been defined for the \"Print Setup...\" entry"}
    $mf add command -label "Print..." -command {error "this is just a demo: no action has been defined for the \"Print...\" entry"}
    $mf add separator
    $mf add command -label "Quit" -command "Exit"

    menu $mo -tearoff 0
    $mo add command -label "Long entry that does nothing"
    if {$tcl_platform(platform) == "macintosh"} {
	set modifier Command
    } elseif {$tcl_platform(platform) == "windows"} {
	set modifier Control
    } else {
	set modifier Meta
    }
    foreach i {A B C D E F} {
	$mo add command -label "Print letter \"$i\"" -underline 14 \
	  -accelerator Meta+$i -command "puts $i" -accelerator $modifier+$i
	bind $w <$modifier-[string tolower $i]> "puts $i"
    }
    if {$isTcl7} {
	eval pack $menubuts -side left -fill y -expand 0
	pack $w.menu -side top -fill x -expand 0
    } else {
	[winfo parent $w.menu] configure -menu $w.menu
    }
    # End of menubar code   
}

set afe_soundfile "" ;# Make sure it exists, so Exit won't crash

proc Exit {{val 0}} {
    # Clean up and exit 
    global afe_soundfile
    if {[file exist $afe_soundfile]} {
	catch "exec /bin/rm $afe_soundfile"
    }
    # Any cleanup required by interface to thislIR
    thislif_cleanup
    exit $val
}

# Build gui (if running under wish)
if {[info vars tk_patchLevel] != ""} {

    # Gray background on mac
    if {$tcl_platform(platform) == "macintosh"} {
	tk_setPalette gray90
    }

    if {![winfo exist .srt]} {
	#set ed980105 [ReadSRT "ed980105"]

	# Setup image bitmaps
	SetupImages

	set rootwin "."
	# Standard idiom to get right prefix from rootwin ("." becomes "")
	regexp {(.*[^.])?} $rootwin w

	# Construct a (dummy) menu bar
	MNUmake $w

	# Label at the top
	label $w.lab -text "ThislIR demo" -font $fontArray(helv24bi) \
		-bg gray95
	pack .lab -padx $pad -pady $pad -side top -fill x -expand 0

	if {$useAudioFrontend} {
	    # Stuff to pass to audio_frontend
	    set afe_frame [frame $w.afe]
	    set afe_root [frame $afe_frame.afe -relief groove -borderwidth 2]
	    set afe_outputHandler "getAudio"
	    set afe_outputHook "afe_setstatus recognizing...; update"

	    pack $afe_root -fill both -expand 0 -side top -padx $pad -pady $pad
	    pack $afe_frame -side left -fill y

	    # Load the audio_frontend display
	    set argv "useloadsave=1 debug=0 samplerate=$samplerate"
	    audio_frontend_init
	    # Hide some buttons we don't want
	    pack forget $quitbutton
	    #pack forget $playbutton
	    # audio_frontend 1.23 now supports play under linux

	    # Make sure pipes are active
	    rdprocEnable

	    # Maybe add an NLP parse display
	    if {$useNLP} {
		set pwin ${afe_frame}.parse
		NLPmake $pwin
		pack $pwin -side bottom -fill both -expand 1 \
			-padx $pad -pady $pad
	    }

	} else {
	    # No audio frontend, so catch the calls to afe_setstatus 
	    proc afe_setstatus state {
		Warn "status: $state"
	    }
	}
	# Build rest of front-end UI elements
	set ww [frame $w.rframe]
	set qrydisp "$ww.qry"
	set resdisp "$ww.res"
	set srtdisp "$ww.srt"
	QRYmake $qrydisp
	pack $qrydisp -side top -fill x -expand 0 -padx $pad -pady $pad
	RESmake $resdisp
	pack $resdisp -side top -fill x -expand 0 -padx $pad -pady $pad
	SRTmake $srtdisp
	pack $srtdisp -fill both -expand 1 -padx $pad -pady $pad
	pack $ww -fill both -side right -expand 1

    }
}
